from typing import List, Optional

import yaml
from clicasso import Slug

from apparatus.core import Command, read_repo_config
from apparatus.config import Config


class Get(Command):
    env: Optional[str] = None
    key: Optional[str] = None

    @classmethod
    def slug(cls) -> Slug:
        return ("config", "get")

    def run(self, config: Config, remainder: List[str]) -> None:
        # pylint: disable-unused-argument
        repo_config = read_repo_config(config, self.env)
        if self.key is not None:
            print(yaml.dump(repo_config[self.key]))
        else:
            print(yaml.dump(repo_config))



from typing import List

from clicasso import Slug
import kubernetes
from kubernetes.client.api_client import ApiClient
from tabulate import tabulate

from apparatus.config import Config
from apparatus.core import Command
from apparatus.utils import maybe


class Resources(Command):
    # pylint: disable=no-member
    @classmethod
    def slug(cls) -> Slug:
        return ("kube", "resources")

    namespace: str

    def run(self, config: Config, remainder: List[str]) -> None:
        records = self._fetch_records(config)
        print(tabulate(records, headers="keys"))

    def _fetch_records(self, config: Config):
        def it(v1):
            result = v1.list_namespaced_pod(namespace=self.namespace)
            for pod in result.items:
                for container in pod.spec.containers:
                    yield {
                        "pod": pod.metadata.name,
                        "container": container.name,
                        "limits.memory": maybe(container.resources.limits, "memory"),
                        "limits.cpu": maybe(container.resources.limits, "cpu"),
                        "requests.memory": maybe(container.resources.requests, "memory"),
                        "requests.cpu": maybe(container.resources.requests, "cpu"),
                    }

        if config.kubernetes is None:
            raise ValueError("No kubernetes config")
        kubernetes.config.load_kube_config(config.kubernetes.config)
        with ApiClient() as api:
            v1 = kubernetes.client.CoreV1Api(api)
            return list(it(v1))



import logging
import os
from typing import List, Optional

import alembic.config
import sqlalchemy
from clicasso import Slug
from pydantic import Field
from retry import retry

from apparatus.config import Config
from apparatus.core import Command, read_repo_config

log = logging.getLogger(__name__)


_RETRY_PARAMS = {
    "exceptions": sqlalchemy.exc.OperationalError,
    "tries": 20,
    "max_delay": 1,
    "delay": 0.2,
    "backoff": 1.5,
}
_ALEMBIC_ARGS = ["upgrade", "head"]


class Migrate(Command):
    init: str = Field(
        "migrations/init.sql",
        description="path to the database initialization SQL script",
        required=False,
    )
    key: str = Field(
        "database",
        description="key for the database configuration in the config map",
        required=False,
    )
    separate: bool = Field(
        False, description="execute each migration in a separate transaction", required=False
    )
    env: Optional[str] = None

    @classmethod
    def slug(cls) -> Slug:
        return ("db", "migrate")

    def run(self, config: Config, remainder: List[str]) -> None:
        config = read_repo_config(config, self.env)
        url = self._database_url(config)
        self._init(url)
        self._run_alembic(url, remainder)

    def _database_url(self, config) -> str:
        config = config[self.key]
        return "{dialect}://{user}:{password}@{host}:{port}/{database}".format(
            dialect=config.get("dialect", "postgresql"),
            user=config["user"],
            password=config["password"],
            host=config["host"],
            port=config["port"],
            database=config["database"],
        )

    def _init(self, url: str) -> None:
        engine = sqlalchemy.create_engine(url)

        @retry(**_RETRY_PARAMS, logger=None)
        def connect():
            return engine.connect()

        log.info("Waiting at most %.02f seconds for connection", self._wait_for())
        conn = None
        try:
            conn = connect()
            if os.path.isfile(self.init):
                with open(self.init) as h:
                    sql = h.read()

                log.info("Applying %s to %r", self.init, engine.url)
                conn.execute(sql)
            else:
                log.info("No file at %s", self.init)
        finally:
            if conn is not None:
                conn.close()

    @classmethod
    def _wait_for(cls) -> float:
        delay = _RETRY_PARAMS["delay"]
        backoff = _RETRY_PARAMS["backoff"]
        max_delay = _RETRY_PARAMS["max_delay"]
        result = 0
        for i in range(_RETRY_PARAMS["tries"] + 1):
            result += min(delay * (backoff ** i), max_delay)
        return result

    def _run_alembic(self, url: str, remainder: List[str]) -> None:
        alembic_args = remainder or _ALEMBIC_ARGS
        alembic_args = ["-x", f"url={url}"] + alembic_args
        if self.separate is True and alembic_args[:2] == ["upgrade", "head"]:
            while True:
                try:
                    alembic.config.main(alembic_args)
                except SystemExit:
                    # The alembic console runner did not produce anymore migrations
                    # TODO: Do not print an alembic ERROR to the log
                    # This is likely difficult to fix when using the default alembic command
                    # line program. Finding a proper solution probably requires understanding
                    # the internals of alembic better.
                    break
        else:
            os.environ["URL"] = url
            alembic.config.main(alembic_args)



from clicasso import CommandParser

from apparatus.commands import COMMANDS
from apparatus.config import read_config


def main():
    parser = CommandParser.from_commands(*COMMANDS)
    command, remainder = parser.parse_known_args()
    config = read_config()
    command.run(config, remainder)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List, Optional

import clicasso
import miniscule

from apparatus.config import Config
from apparatus.exceptions import GenericError


class Command(clicasso.BaseCommand, ABC):
    @abstractmethod
    def run(self, config: Config, remainder: List[str]) -> None:
        pass


def read_repo_config(config: Config, env: Optional[str]):
    if config.repository is None:
        raise GenericError("No repository config")
    if env is not None:
        path = config.repository.config.format(env=env)
        return miniscule.read_config(path)
    return miniscule.read_config()



from typing import Optional
from pydantic import BaseModel


class Repository(BaseModel):
    config: str


class Kubernetes(BaseModel):
    config: Optional[str] = None


class Config(BaseModel):
    repository: Optional[Repository] = None
    kubernetes: Optional[Kubernetes] = None



import os
from pathlib import Path
from typing import Iterator, Optional


def maybe(m, key):
    if m is None:
        return None
    return m.get(key)


def merge(acc, piece):
    if not isinstance(acc, dict):
        return piece
    for (key, value) in piece.items():
        if key not in acc:
            acc[key] = value
        else:
            acc[key] = merge(acc[key], piece[key])
    return acc


def parents(path: Optional[Path] = None) -> Iterator[Path]:
    path = path or Path(os.getcwd())
    parts = path.parts
    current = Path(parts[0])
    yield current
    for part in parts[1:]:
        current = current.joinpath(part)
        yield current



# pylint: disable=unused-import
from typing import List

from clicasso import Command

COMMANDS: List[Command] = []
_EXCEPTIONS = (ImportError, ModuleNotFoundError)


try:
    from apparatus.commands import db

    COMMANDS.extend(db.COMMANDS)
except _EXCEPTIONS as exc:
    pass

try:
    from apparatus.commands import config

    COMMANDS.extend(config.COMMANDS)
except _EXCEPTIONS as exc:
    pass


try:
    from apparatus.commands import kube

    COMMANDS.extend(kube.COMMANDS)
except _EXCEPTIONS as exc:
    pass



from pathlib import Path
from typing import Any, Iterator, Optional

import miniscule

from apparatus.utils import merge, parents
from apparatus.config.core import Config

Object = Any


def configs(path: Optional[Path] = None) -> Iterator[Object]:
    for parent in parents(path):
        try:
            with open(parent.joinpath(".apparatus.yaml"), "r") as stream:
                obj = miniscule.load_config(stream)
                if obj is not None:
                    yield obj
        except FileNotFoundError:
            pass


def read_config() -> Object:
    acc: Object = {}
    for config in configs():
        acc = merge(acc, config)
    return Config(**acc)


if __name__ == "__main__":
    print(read_config())
